# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Data.py - Functions to load data from Matlab files and to 
    interpolate the frequency and magnitude of the maxima.

"""

__all__ = ['LoadMatlabData', 'InterpolateMaxima']

import scipy.io
from scipy import optimize, linalg
import numpy
from numpy import sign, diff

from wiat.util import calculate_interval_in_cents

def load_from_mat(names, filename):
    """ Load data from a matlab file. All variable name in the names tuple are 
        fetched and returned .

        Example::

        >>> import os
        >>> import numpy
        >>> import wiat
        >>> f,z = wiat.Data.load_from_mat(('freqs','z'),
	...    os.path.join(wiat.MEASUREMENTDIR,'20080412Ztoneholeless.mat'))
	>>> numpy.size(f) == numpy.size(z)
	True
	

        @param names: A tuple containing variable names to get from the Matlab
            file.
	@param filename: the path to the Matlab file
	@return: a list of the extracted arrays. None if the name do not exist.
    """

    try:
        data = scipy.io.loadmat(filename)
    except IOError, arg:
        print arg
        raise IOError(arg)

    ret = []
    for name in names:
        if not data.has_key(name):
            print 'Warning: no such variable %s in Matlab file %s'\
                    %(name,filename)
        ret.append(data.get(name))
    return ret

class __Parabola:
    " Class for fiting an equation of the form M{y = ax**2 + bx + c} "
    def __init__(self):
        self.n = 0
        
    def __call__(self, f):
        c = self.C
        return c[0]*f**2+c[1]*f+c[2]
    
    def fit_points(self, px, py):
        """ Fit the parabola passing through 3 points with 3 equations

            M{y1 = a * x1**2 + b*x1 + c}
            M{y2 = a * x2**2 + b*x2 + c}
            M{y3 = a * x3**2 + b*x3 + c}

        """
        self.px = px
        self.py = py
        B = py
        tmp = numpy.array([px**2, px, [1.,1.,1.]])
        A = numpy.array([tmp[:,0], tmp[:,1], tmp[:,2]])
        self.C = scipy.linalg.solve(A, B)

    def lowf(self):
        return self.px[0]

    def highf(self):
        return self.px[2]

    def xc(self):
        """ position of vertical axis of symmetry: h = -b / 2a """
        return - self.C[1] / (2*self.C[0])

    def yc(self):
        """ value of parabola on axis of symmetry: k = (4ac - b**2) / 4a """
        return (4*self.C[0]*self.C[2] - self.C[1]**2) / (4 * self.C[0])
    
def interpolate_maxima(f, Z, number_of_maxima=None, fmin=100.):
    """ Find the maxima in the data and interpolate with Parbola to find the 
        frequency and amplitude of the maxima.

        Results that are noisy are likely to pose problems to the algorithm.

        @param fmin: minimal frequency we are interested in, useful to avoid 
            low frequency noise to infect the results.
    """

    derivative = diff(abs(Z))
    signarray = sign(derivative)
    crossing  = diff(signarray)
    indices   = crossing.nonzero()

    maximalist = []

    for i in indices[0]:
        if f[i] < fmin:
            continue

       # when there is noise in the data, some maxima appears at various 
       # points in the curve.
       # TODO: find a method to eliminate those false maxima

            # in order to eliminate them, we can check various criteria.
            # generally, the derivative of the phase is large at the maxima
        #if abs(angle(Z[i])-angle(Z[i+1]))/(f[i+1]-f[1]) < 0.001:
        #    print 'skipping f = %f'%f[i]
        #    continue

        # minima are much smaller than 1, maxima are much larger
        if abs(Z[i]) > 1.:
            # find the indice of the maxima around i
            while True:
                if abs(Z[i-1]) > abs(Z[i]):
                    i = i - 1
                elif abs(Z[i+1]) > abs(Z[i]):
                    i = i + 1
                else:
                    break
                    
            P = __Parabola()
            P.fit_points(f[i-1:i+2], abs(Z[i-1:i+2]))

            # if the new maxima is closer than a semitone to the previous one
            # it is likely a duplicate due to noise... the highest one is kept
            if maximalist != []:
                if abs(calculate_interval_in_cents(P.xc(), maximalist[-1][0]))\
                   < 300.:
                    if P.yc() > maximalist[-1][1]:
                        maximalist.pop()
                    else:
                        continue
                        
            #parabolas.append([i, P])                
            maximalist.append( ( P.xc(), P.yc() ) )
    return maximalist


# Code to be adapted
#
#def fitfunc(p,f):
#    return  ( p[0] / (1 + p[1]*(f - p[2])**2 ) )
#
#def errfunc(p,f,y):
#    ff = fitfunc(p,f)
#    return (ff - y)
#
#def fitLorentzPeak(self):
#    l = self.listMaxima()
#    print "Fitting peaks for %s" % self.desc.object
#
#    self.LorentzPeaks = []
#        
#    for k in arange(0,len(l),1):
#        i = self.maximalist[k]
#        # test if it is a maximum or minimum and proceed for maximum only
#        #print abs(self.Z[i])
#        if abs(self.Z[i]) > 1.5:
#            #print 'working on maxima no. %d'%k
#            # p0 contains three value, the peak height, a value related to 
             # peak width, and frequency
#            A = abs(self.Z[i])
#            p0 = [A, 0.05, self.f[i]]  # Initial guess for the parameters
#            deltaf = self.f[1]-self.f[0]
#            ni = int(self.f[i]*(pow(2.,1./48.)-1.)/deltaf) 
                # use a small interval around the center
#            if ni <= 1:
#                ni = 2
#                    
#            nc = self.maximalist[k]
#            #print ni, nc
#            range = arange(nc-ni,nc+ni+1,1)
#
#            #print 'optimizing with points', 
              #(self.f[range], abs(self.Z[range]))
#                                              
#            p1,success = scipy.optimize.leastsq(errfunc, p0[:],
#args = (self.f[range], abs(self.Z[range])))
#            #print p0, p1, success
#
#            f1 = p1[2]
#            nc = int(round((f1-self.f[0])/deltaf))
#            range = arange(nc-ni,nc+ni,1)
#            p1,success = scipy.optimize.leastsq(errfunc, p0[:], 
#args = (self.f[range], abs(self.Z[range])))
#            #print p0, p1, success
#            # TODO: recalculate the center indice from p1[2] 
#(the frequency) and reoptimize
#               
#            fmax=self.f[nc+ni]
#            fmin=self.f[nc-ni]
#            if success and p1[2] > 0.:
#                self.LorentzPeaks.append([p1,fmin,fmax])
#            else:
#                print 'maxima rejected'
