# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Functions.py - Utility functions for the multimodal solver.

    In order to solve a complex matricial system of ordinary differential
    equations, it must be convert to a real system of equation. Two functions
    are used to convert back and forth from the matrix notation to the
    vectorial form: flat2matrix and matrix2flat.

"""

import numpy

def flat2matrix(flat_matrix,N):
    """ Convert a flat vector (obtained with matrix2flat) to a matrix of 
        complex numbers

        @param flat_matrix: a flat vector representation of a matrix
        @param N: the size of the square matrix
    """
    [flat_matrix_real, flat_matrix_imag] = numpy.split(flat_matrix,2)
    matrix = flat_matrix_real.reshape(N,N) + 1j*flat_matrix_imag.reshape(N,N)
    return matrix

def matrix2flat(matrix):
    """ Transform a matrix of complex numbers in a vector of real number 
        consisting of a flatten version of the original.

        Example

        >>> import numpy
        >>> import wiat
        >>> mat = numpy.array([[complex(1.,1.), complex(2.,2.)],
        ...                    [complex(3.,3.), complex(4.,4.)]])
        >>> fmat = wiat.MM.Functions.matrix2flat(mat)
        >>> print mat
        [[ 1.+1.j  2.+2.j]
         [ 3.+3.j  4.+4.j]]
        >>> print fmat
        [ 1.  2.  3.  4.  1.  2.  3.  4.]

    """
    flat_matrix = numpy.concatenate((numpy.real(matrix.flatten()), 
                                     numpy.imag(matrix.flatten())))
    return flat_matrix
