# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Radiation.py - Multimodal radiation matrix implementations.

"""

import os
import numpy
from numpy import sin, cosh, sqrt, real, imag, pi
import scipy
import scipy.integrate
import scipy.interpolate
from scipy.interpolate import UnivariateSpline
import scipy.io
import scipy.special
from scipy.special import j0,j1,jn_zeros,jvp

from Functions import flat2matrix, matrix2flat

class RadiationFlanged:
    """ Class implementing the Zorumski multimodal radiation impedance matrix.
    
        [1] N. Amir, V. Pagneux, and J. Kergomard,
        "A study of wave propagation in varying cross-section waveguides by 
        modal decomposition. Part II. Results,"
        The Journal of the Acoustical Society of America,  
        vol. 101, May. 1997, pp. 2504-2517.

        [2] W.E. Zorumski, "Generalized radiation impedances and reflection 
        coefficients of circular and annular ducts,"
        The Journal of the Acoustical Society of America,  
        vol. 54, Dec. 1973, pp. 1667-1673.

        [3] Jonathan Kemp, "Theoretical and experimental study of wave 
        propagation in brass musical instruments",
        Ph.D. Thesis, University of Edinburgh, 2002        
        
    """
    N = 15
    
    def __init__(self,diameter,number_of_modes = N, do_interpolation = True):
        self.radius = diameter / 2.
        self.number_of_modes = number_of_modes

        assert number_of_modes >=1, 'Minimal number of modes is 1'
        if number_of_modes == 1:
            self.gamma = [0.]
        else:
            self.gamma = numpy.concatenate(
                    ([0.],jn_zeros(1,self.number_of_modes-1)))

        # if there is a data file, load it and do interpolation
        self.do_interpolation = do_interpolation
        self.backup_filename = os.path.join(os.path.dirname(__file__),'Znm')
        if do_interpolation:
            try:
                self.data_table = scipy.io.loadmat(self.backup_filename)
                N = int(self.data_table['number_of_modes'])
                Znm = numpy.reshape(self.data_table['Z_nm'],
                                    (N,N,self.data_table['length_ka']))
                self.data_table['Z_nm'] = Znm
                #print 'Creating interpolation splines for the multimodal 
                #radiation.'
                self.spl_real = []
                self.spl_imag = []
                for i in range(N):
                    self.spl_real.append([])
                    self.spl_imag.append([])
                    for j in range(N):
                        #print self.data_table['ka']
                        #print numpy.real(self.data_table['Z_nm'][i][j])
                        self.spl_real[i].append(
                                UnivariateSpline(self.data_table['ka'],
                                    numpy.real(self.data_table['Z_nm'][i][j]),
                                    s=0.0000001))
                        self.spl_imag[i].append(
                                UnivariateSpline(self.data_table['ka'],
                                    numpy.imag(self.data_table['Z_nm'][i][j]),
                                    s=0.0000001))
            
            except IOError:
                #print arg
                print 'Calculating radiation data table for interpolation.'
                # if there is no data file, create it
                ka = numpy.arange(0.,15.,0.1)
                self.do_interpolation = False
                self.data_table = {'number_of_modes': self.number_of_modes,
                                   'length_ka': numpy.size(ka,-1),
                                   'ka': ka,
                                   'Z_nm': self.calculateZnm(ka/self.radius)}
                self.do_interpolation = True
                scipy.io.savemat(self.backup_filename, self.data_table)
        
    def __call__(self,N,k):
        assert N <= self.number_of_modes,\
            'number of modes exceed radiation impedance maximum'
        Z = numpy.empty((N,N),dtype=complex)
        for i in range(N):
            for j in range(N):
                    Z[i][j] = self.Z_ij(i,j,k)
        return matrix2flat(Z)
    
    def Phi(self,r,i):
        return j0(self.gamma[i]*r/self.radius)/j0(self.gamma[i])

    def D(self,tau,i,k):
        """ from [3] eq. 4.1 """
        lambdan = self.gamma[i]/(k*self.radius)
        if i == 0 and tau == 0: return 0.
        return -sqrt(2.)*tau*j1(tau*k*self.radius) / (lambdan**2-tau**2)

    def integrand_Z_ij_real(self,phi,i,j,k):
        return sin(phi) * self.D(sin(phi),i,k) * self.D(sin(phi),j,k)

    def integrand_Z_ij_imag(self,xhi,i,j,k):
        return cosh(xhi) * self.D(cosh(xhi),i,k) * self.D(cosh(xhi),j,k)
            
    def Z_ij(self,i,j,k):
        #print i,j
        if self.do_interpolation:
            ka = k * self.radius 
            if ka >= self.data_table['ka'][-1]:
                print 'Warning: value of ka not in radiation matrix table'
            Z_ij = complex(self.spl_real[i][j](ka)[0],
                           self.spl_imag[i][j](ka)[0])
        else:
            if k == 0.:
                Z_ij = complex(0.,0.)
            else:
                realpart = scipy.integrate.quad(self.integrand_Z_ij_real,
                        0., numpy.pi/2,args=(i,j,k),limit=5000)
                imagpart = scipy.integrate.quad(self.integrand_Z_ij_imag, 
                        0., 100, args=(i,j,k),limit=5000)

                # TODO- Check the sign of the imaginary part 
                # (kemp is positive, Zorumski negative)
                # Antoine Lefebvre (2008) - It seems that the negative sign 
                # is correct,
                # from comparison with plane wave models
                Z_ij =  complex(realpart[0], -imagpart[0])

        return Z_ij
    
    def calculateZnm(self,ks,number_of_modes=N):        
        Zs = numpy.empty((number_of_modes, number_of_modes,len(ks)),
                         dtype=complex)
        l = 0
        print "Calculating flanged modal radiation matrix..."
        for k in ks:
            for i in range(number_of_modes):
                for j in range(number_of_modes):
                    Zs[i][j][l] = self.Z_ij(i,j,k)
            l = l+1
        print "done."

        return Zs
        


# this is not correct
# a formulation for the multimodal radiation impedance have to be found
# maybe in
#
# Wang, K. S., & Tszeng, T. C. (1984).
# Propagation and radiation of sound in a finite length duct.
# Journal of Sound and Vibration, 93(1), 57-79.
class RadiationUnflanged:
    def __init__(self,radius):
        self.a = radius

    def __call__(self,N,k):
        ka = k * self.a
        R0 = (1 + 0.2*ka - 0.084*ka**2)/(1 + 0.2*ka + (0.5 - 0.084)*ka**2)
        delta = 0.6113 * ((1+0.044*ka**2)/(1+0.19*ka**2) - 0.02*(numpy.sin(2*ka))**2 )
        R = - R0 * numpy.exp( -2 * 1j * ka * delta)
        Zl = (1+R)/(1-R)
        M = numpy.diag(Zl*numpy.ones(N))
        return matrix2flat(M)

    

class RadiationIdealOpen:
    pass

class CharacteristicImpedance:
    pass

def AdmittanceClosed(N,k):
    """ Rigid wall admittance matrix is zero (N x N)
        they are complex valued so we should return a flat vector of 2*N**2 values
    """
    return numpy.zeros(2*N**2)

#    Z = numpy.diag( self.Zc(0.,k)/10. , 0)
#        return numpy.concatenate( (numpy.real(Z.flatten()), numpy.imag(Z.flatten()) ))
#    pass

