# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Solver.py - Multimodal wave propagation solver.

    This is an implementation of the theory presented in references [1] and [2]
    with boundary layer loss model from reference [3].

    References:

        1. V. Pagneux, N. Amir, and J. Kergomard,
           "A study of wave propagation in varying cross-section waveguides by
	   modal decomposition. Part I. Theory and validation,"
           J. Acoust. Soc. Am.,  vol. 100, 1996, pp. 2034-2048.

        2. Amir, N., Pagneux, V., & Kergomard, J. (1997). 
           A study of wave propagation in varying cross-section waveguides by 
           modal decomposition. Part II. Results. 
           The Journal of the Acoustical Society of America, 101(5), 2504-2517.
           
        3. Bruneau, A. M., Bruneau, M., Herzog, P., & Kergomard, J. (1987).
           "Boundary layer attenuation of higher order modes in waveguides."
           Journal of Sound and Vibration, 119(1), 15-27.


    Example::

    >>> import wiat
    >>> din = 0.02
    >>> dout = 0.025
    >>> f = numpy.arange(50.,500.,10.)
    >>> air = wiat.util.air_properties(25.)
    >>> c = wiat.MM.Waveguides.QuarterCircleHorn(din, dout)
    >>> rad = wiat.MM.Radiation.RadiationFlanged(dout)
    >>> Z5 = wiat.MM.Solver.SolveImpedance(c,rad,f,air,5)

"""

__all__ = ['SolveImpedance', 'SolveAdmittance']

import time

import numpy
from numpy import sin, cosh, sqrt, real, imag, pi, dot, transpose
import scipy.special
from scipy.special import j0,j1,jn_zeros,jvp

from wiat.util import air_properties
from Functions import flat2matrix, matrix2flat

def calculate_gamma(N):
    return numpy.concatenate(([0.],scipy.special.jn_zeros(1,N)))

def _Q(N):
    """ Q is the constant matrix defined in reference 1, equation 28 
 
        @param N: number of modes 
    """
    gamma = calculate_gamma(N) 
    Q = []
    for n in range(N):
        Q.append([])
        gamman2 = gamma[n]**2
        for m in range(N):
            if m==n:
                Q[n].append(0.)
            else:
                gammam2 = gamma[m]**2
                Q[n].append(gammam2 / (gammam2 - gamman2))
    return numpy.array(Q)

def _K(N,a,k,air):
    """ Multimodal propagation constant 
        (with boundary layer thermo-viscous losses)

        See [2], equations 5 and 6 as well [3]

    """
    gamma = calculate_gamma(N)
    lv = air['mu']/(air['rho']*air['c'])
    lt = lv/air['nu']**2
    epsilonv = complex(1.,1.) * sqrt(k/2) * sqrt(lv)
    epsilont = complex(1.,1.) * sqrt(k/2) * (air['gamma']-1)*sqrt(lt)
    K = []
    for i in range(N):
        alpha = gamma[i]/a
        epsiloni = (1. - alpha**2/k**2)*epsilonv + epsilont
        K.append( k**2 - alpha**2 + (2j*k/a)*
                  (numpy.imag(epsiloni) -1j*numpy.real(epsiloni)))
    return K

def _Zprime(Zflat,z,geometry,k,air,N,Q):
    """ Matrical Riccati equation for the generalized impedance matrix Z
    
        This is an implementation of equation 32 of reference [1], which is 
        the same as equation 1 of reference [2].

    """
    Z = flat2matrix(Zflat,N)
    Zc = air['rho']*air['c']/geometry.S(z)
    K = numpy.diag(_K(N,geometry.radius(z),k,air), 0)
    tmp = dot(K,Z)
    tmp = dot(Z,tmp)/(k*Zc)
    tmp2 = numpy.diag(numpy.ones((N))*k*Zc, 0)
    Zp = 1j*(-tmp2+tmp)+geometry.Sp_S(z)*(dot(Q,Z) + dot(Z,transpose(Q)))
    return matrix2flat(Zp)


def SolveImpedance(geometry,Zr,freqs,air,N):
    freqs = numpy.asarray(freqs).flatten()

    zf = geometry.get_length()
    zs = numpy.array([0., zf])
    Z00 = []
    Zc = air['rho']*air['c']/geometry.S(zf)
    Zcr = air['rho']*air['c']/geometry.S(0.)
    print "Integrating system for %d modes and %d frequencies..."%(N,len(freqs))
    initialtime = time.clock()
    Q = _Q(N)
    for f in freqs:

        # when frequency is 0, it is a flow problem and the method can't work
        # we append a null impedance in that case
        if f == 0.:
            Z00.append(complex(0.,0.))
            continue

        previoustime = time.clock()
        k = 2*numpy.pi*f/air['c']
        Zflat,infodict = scipy.integrate.odeint(
            _Zprime, Zcr*Zr(N,k), zs, (geometry,k,air,N,Q),
            None, full_output=1, mxstep=15000,
            rtol=1e-5, atol=1e-10)
        Z = flat2matrix(Zflat[1],N)
        Z00.append(Z[0][0]/Zc)
        print "% 7.2f %.2f seconds"%(f, time.clock() - previoustime)
    print "...done in %.2f seconds"%(time.clock() - initialtime)
    return numpy.array(Z00)

def _Yprime(Yflat,z,geometry,k,air,N,Q):
    """ Riccati equation for admittance
    
        Implementation of equation 4 of reference [2].
    """
    Y = flat2matrix(Yflat, N)
    kZc = k * air['rho'] * air['c'] / geometry.S(z)
    K = numpy.diag(_K(N,geometry.radius(z),k,air), 0)
    Yp = -1j*K/(kZc) - geometry.Sp_S(z)*(dot(transpose(Q),Y) + dot(Y,Q)) +\
            1j*kZc*dot(Y,Y)
    return matrix2flat(Yp)

def SolveAdmittance(geometry,Yr,freqs,air,N):
    freqs = numpy.asarray(freqs).flatten()

    zf = geometry.L
    zs = numpy.array([0., zf])
    Y00 = []
    Yc = geometry.S(zf)/(air['rho']*air['c'])
    print "Integrating system for %d modes and %d frequencies..."%(N,len(freqs))
    initialtime = time.clock()
    Q = _Q(N)
    for f in freqs:
        previoustime = time.clock()
        k = 2*numpy.pi*f/air['c']
        Yflat,infodict = scipy.integrate.odeint(
                _Yprime, Yr(N,k),zs,(geometry,k,air,N,Q),
                None,full_output=1,mxstep=15000,
                rtol=1e-5,atol=1e-10)
        Y = flat2matrix(Yflat[1],N)
        Y00.append(Y[0][0]/Yc)
        print "% 7.2f %.2f seconds"%(f, time.clock() - previoustime)
    print "...done in %.2f seconds"%(time.clock() - initialtime)
    return numpy.array(Y00)

# to be done... 
#def CalculateMaxima(Z,f,solver,inst,rad,number_of_maxima=None,**kwargs):
#    derivative = numpy.diff(numpy.abs(Z))
#    signarray = numpy.sign(derivative)
#    crossing  = numpy.diff(signarray)
#    indices   = crossing.nonzero()
#    maximalist = []
#    for i in indices[0]:
#        # do not take into account the minimum        
#        if crossing[i] > 0:
#            continue
#        if len(maximalist) == number_of_maxima:
#            break
#        f0 = f[i-1]
#        f1 = f[i+1]
#        if kwargs.get('Y'):
#            results = scipy.optimize.brent(lambda f: numpy.abs(solver.Ysolve(f,inst,rad)),brack=(f0,f1),full_output=True)
#        else:
#            results = scipy.optimize.brent(lambda f: 1./numpy.abs(solver.Zsolve(f,inst,rad)),brack=(f0,f1),full_output=True)
#        fres = results[0][0]
#        zres = 1./results[1][0]
#        #fres = scipy.optimize.brentq(lambda f: numpy.imag(solver.Zsolve(numpy.array([f]),inst,rad)),f0,f1)
#        if zres > 1.: # it is a maxima and not a minima
#            maximalist.append((fres, zres))
#        
#    return maximalist
