# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Waveguides.py - Definition of waveguides for use with the multimodal
    wave propagation solver.

"""

import math
import numpy
from numpy import sin, cos, cosh, arcsin, sqrt, real, imag, pi, log10

class WaveguideProfile:
    def __init__(self):
        print "test"

    def assertZ(self,z):
        pass
        #assert 0. <= z <= self.L, "position out of range"

    def get_length(self):
        return self.L

    def radius(self,z):
        " to be defined be derived class "
        pass

    def radiusp(self,z):
        " to be defined by derived class "

    def S(self,z):
        " return the area of the axisymmetric duct from free end to source "
        self.assertZ(z)
        return numpy.pi * self.radius(z)**2

    def Sp(self,z):
        self.assertZ(z)
        return 2 * numpy.pi * self.radius(z) * self.radiusp(z)

    def Sp_S(self,z):
        self.assertZ(z)
        return 2 * self.radiusp(z) / self.radius(z)
    
class Cone(WaveguideProfile):
    def __init__(self,d0,de,L):
        self.a0 = d0/2
        self.ae = de/2
        self.L = L
    def radius(self,z):
        self.assertZ(z)
        return self.ae + (self.a0-self.ae)*z/self.L
    
    def radiusp(self,z):
        self.assertZ(z)
        return (self.a0-self.ae)/self.L

class Cylinder(WaveguideProfile):
    def __init__(self,d,L):
        self.a = d/2
        self.L = L
    def radius(self,z):
        self.assertZ(z)
        return self.a
    
    def radiusp(self,z):
        self.assertZ(z)
        return 0.

class QuarterCircleHorn(WaveguideProfile):
    def __init__(self,d0,de,theta1=0.,theta2=numpy.pi/2.):
        """
	d1: input diameter of the horn
	d2: output diameter of the horn
        theta1: input angle of the horn (radians)
	theta2: output angle of the horn (radians)

        a: input radius of the horn
        r: radius of the quarter circle
        """
        self.d0 = d0
        self.de = de
        self.theta1 = theta1
        self.theta2 = theta2

        self.a0 = d0/2.
        self.r_horn = (de-d0) / (2. * (cos(theta1)-cos(theta2)))
        self.L = self.r_horn * (sin(theta2)-sin(theta1))
#        if theta2 == numpy.pi/2.:
#            theta2 = theta2 - 0.000001 

        self.ze = self.r_horn * sin(theta2)
        self.y0 = self.r_horn * cos(theta1) 

    def __str__(self):
        return "QuarterCircleHorn(d0=%.4f,de=%.4f,theta1=%.2f,theta2=%.2f)"%\
                (self.d0,self.de,self.theta1,self.theta2)
    def radius(self,z):
        self.assertZ(z)
        theta = arcsin( (self.ze - z) / self.r_horn)
        return self.a0 + self.y0 - self.r_horn * cos(theta)

    def radiusp(self,z):
        """ derivative of the radius along z
            the 0.000001 value was added to avoid infinite derivatives
        """
        self.assertZ(z)
        arg = (self.ze - z) / self.r_horn
        theta = arcsin( arg )
        if (1. - arg) < 1e-15:
            arg = 1. - 1.e-15
        dtheta = -1. / (self.r_horn * sqrt(1.-arg**2))
        return  self.r_horn * sin(theta) * dtheta

class BesselHorn(WaveguideProfile):
    def __init__(self,d1,d2,L,alpha):
        self.L = L
        self.alpha = alpha
        self.B = 10**(log10(L) - log10(d1**-1-d2**-1))
        self.z0 = 10**( alpha**-1 * log10(self.B/d2) )

    def radius(self,z):
        return 0.5 * self.B / (self.L - z + self.z0)**self.alpha

    def radiusp(self,z):
        return 0.5 * self.B * self.alpha / \
                (self.L - z + self.z0)**(self.alpha+1.)

#class ExponentialHorn(WaveguideProfile):
#    def __init__(self,d1,d2,L,theta1):
#        self.L = L
#        self.a1 = d1/2.
#	self.a2 = d2/2.
#        self.m = numpy.tan(math.radians(theta1/2.)) / self.a1
#
#    def radius(self,z):
#        return numpy.sqrt( self.S0 * numpy.exp(self.m * z) / numpy.pi )
#
#    def radiusp(self,z):
#        pass
#    
