# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Inst.py - This file contains the class defining an Instrument and the
    functions CalculateImpedance and CalculateMaxima.

    The CalculateImpedance returns the impedance of an instrument for each
    requested frequency.

    The CalculateMaxima seach for the exact location of the maxima of the
    instrument and return a list of (frequency,magnitude) tuples.
	
"""

__all__ = ['Instrument','CalculateImpedance','CalculateMaxima']

import numpy
import scipy.optimize

from Waveguides import ToneHole

def create_identity_matrix_array(n):
    """ Returns an array of 2X2 identity matrices for uses in the transmission
        matrix calculation method.
    """
    M = numpy.array([[numpy.ones(n),numpy.zeros(n)],
	             [numpy.zeros(n),numpy.ones(n)]])
    return M

def multiply_matrix_arrays(TM1,TM2):
    """ Multiplication of two arrays of 2X2 matrices.

        Example::
        
        >>> TM1 = create_identity_matrix_array(10)
        >>> TM2 = multiply_matrix_arrays(TM1,TM1)
        >>> TM1 == TM2
        array([[[ True,  True,  True,  True,  True,  True,  True,  True,  True,
                  True],
                [ True,  True,  True,  True,  True,  True,  True,  True,  True,
                  True]],
        <BLANKLINE>
               [[ True,  True,  True,  True,  True,  True,  True,  True,  True,
                  True],
                [ True,  True,  True,  True,  True,  True,  True,  True,  True,
                  True]]], dtype=bool)
        >>>

    """
    n = range(numpy.shape(TM1)[-1])
    for i in n:
        if i == 0:
            R = numpy.dot(TM1[...,i],TM2[...,i])
        else:
            T = numpy.dot(TM1[...,i],TM2[...,i])
            R = numpy.dstack((R, T))
    if numpy.size(n) == 1:
        return numpy.resize(R,(2,numpy.shape(TM2)[1],1))
    else:
        return R

	
class Instrument:
    """ Class used to descibed the geometry of a musical instrument.
    """
    def __init__(self, name, sections, radiation,
                 toneholes=None, fingerings=None):
        """ Instrument class

            Example::

            >>> import wiat
            >>> from wiat.TM import *
            >>> I1 = Instrument("Cylinder with one tone hole",
            ...                [Cylinder(0.014,0.583)], UnflangedOpenEnd(0.014),
            ...                [ToneHole(0.0117,0.0043,0.474-0.0117/2),
            ...                 ToneHole(0.0112,0.0043,0.505-0.0112/2)],
            ...                {'D':[1,1], 'C sharp':[0,1], 'C':[0, 0]})

            By default, all toneholes of the instrument are supposed close. Then
            the fingering can be set with the set_fingering method.


            @param name: the name of the instrument
            @type name: string
            @param sections: a list of Waveguides
            @type sections: list
            @param radiation: a radiation condition
            @type radiation: an instance of a radiation class (see Radiation.py)
            @param toneholes: optional list of ToneHole 
            @type toneholes: list
            @param fingerings: an optional fingering dictionary
                The fingering dictionary should contains lists of open and 
                closed toneholes. The length of the lists should be the same 
                as the number of toneholes.
            @type fingerings: dictionary

	"""
        self.name = name
        assert type(sections) is type([]), \
            'The second argument should be a list.'
        self.sections = sections
        self.radiation = radiation
        self.toneholes = toneholes

        # self.fingerings is a dictionary
        # self.fingering is a list

        self.fingerings = fingerings
        if toneholes is not None:
            self.fingering = [0]*len(toneholes) # default closed toneholes

    def __str__(self):
        string = '==============================================\n'
        string += "Instrument: %s\n\n"%self.name

        string += "list of sections\n"
        string += "----------------\n"
        for s in self.sections:
            string = string + str(s) + '\n'
        
        string += '\nradiation\n'
        string += '----------\n'
        string += str(self.radiation)+'\n'
        string += '==============================================\n'
        return string

    def add_tonehole(self,tonehole,state=1):
        """ Add a new tonehole to the instrument.

        This method is useful in the design stage of an instrument.

        @param tonehole: the new tonehole to be added.
        @type tonehole: ToneHole
        @param state: if the tonehole is to be added in the open (1) or 
        closed (0) state.
        """
        if not isinstance(tonehole,ToneHole):
            raise TypeError('Expecting a tonehole')

        if self.toneholes is None:
            self.toneholes =  [tonehole]
            self.fingering =  [state]
        else:
            self.toneholes.append(tonehole)
            self.fingering.append(state)

    def get_fingerings(self):
        " Returns the fingering dictionary of the instrument "
        return self.fingerings

    def set_fingering(self,fingering):
        """ Set the fingering of the instrument. 

	    The fingering name must be a key of the fingering dictionary.

	"""
        if self.fingerings is not None:
            if self.fingerings.has_key(fingering):
                self.fingering = self.fingerings[fingering]
            else:
                print "Unknown fingering: %s"%fingering
        else:
            print "Can't set fingering, there is no fingering dictionary"
        return self

    def set_register_key(self,state=1):
        " Allow setting the state of the register key "
        # this function assumes that the first tonehole of the list is the
        # register hole
        # this is not always true, and some instruments has many register holes
        # a solution must be found
        try:
            self.fingering[0] = state
        except AttributeError:
            print "no fingering list available"
        return self

    def get_number_of_sections(self):
        " Returns the number of sections the instrument is made off "
        return numpy.size(self.sections)

    def calculate_total_length(self):
        " Return the physical length of the instrument. "
        L = 0
        for section in self.sections:
            L = L + section.getLength()
        return L

    def find_toneholes_in_range(self,xi,xf):
        """ Return a list of the toneholes located in a specific section of the 
            instrument 
        """
        found_toneholes = []
        if self.toneholes is not None:
             found_toneholes = [(i,tonehole) for (i,tonehole) in \
                                enumerate(self.toneholes) if \
                                xi < tonehole.get_position() <= xf]
        return found_toneholes

def CalculateImpedance(instrument,freqs,T=25.):
    """ Calculate the input impedance of the instrument.

        This function performs the transfer matrix multiplication.
        
        The frequency vector f must be sufficiently fine for the impedance to
        be well represented, otherwise, the CalculateMaxima functions won't be
        able to find the maxima. A good vector is:

        >>> freqs = numpy.arange(50,5000,5,dtype=float)

        @param instrument: an Instrument instance
        @param freqs: frequency vector
        @param T: the temperature

    """
    
    # make sur f is an array of dimension 1
    f = numpy.asarray(freqs).flatten()
    
    # to calculate the input impedance, we must multiply all the 
    # transmission matrices
    # the tonehole transmission matrices must be inserted into the bore sections
    # bore sections must therefore be splitted
    
    xi = 0.
    TM = create_identity_matrix_array(numpy.size(f))

    #print "Calculating input impedance of the instrument: %s..."%instrument.name
    for i,s in enumerate(instrument.sections):
        #print 'processing section %d'%i
        xf = xi + s.get_length()
        found_toneholes = instrument.find_toneholes_in_range(xi, xf)
        #print "Number of toneholes in section %d: %d"%(i,len(SubToneHolesList))

        x0 = xi
        for (j,tonehole) in found_toneholes:
            #print 'Processing tonehole %d'%j
            x1 = tonehole.get_position()
            TM = multiply_matrix_arrays(TM, s.subTM(f,T,x0-xi,x1-xi))
            try:
                TMh = tonehole.TM(f,s.radius(x1-xi),T,instrument.fingering[j])
            except:
                print "Error processing tone hole no. %d"%j
                raise

            TM = multiply_matrix_arrays(TM, TMh)
            x0 = x1 # last tonehole become first coordinate of new sub section

        # last subsection 
        # (if there were no subsections, this will be the complete section
        TM = multiply_matrix_arrays(TM, s.subTM(f,T,x0-xi,xf-xi))
        xi = xf

    TM = multiply_matrix_arrays(TM, instrument.radiation.TM(f,T))
    Z = TM[0]/TM[1]
    #print "...done"
    return Z[0]

def CalculateMaxima(Z,inst,freqs,T,number_of_maxima=None):
    """ Calculate the exact location of the impedance maxima of an Instrument.

        It uses an impedance vector Z, previously calculated with the 
        CalculateImpedance function, to determine the approximate location
        of the maxima and then, uses an optimization algorithm to find their 
        exact location.

        @param Z: impedance vector
        @param inst: Instrument instance
        @param freqs: frequency vector
        @param T: temperature
        @param number_of_maxima: optional argument to limit the number of 
            maxima that are returned
        @return: a list of tuples containing frequency and magnitude of each
           impedance maxima.
    """

    derivative = numpy.diff(numpy.abs(Z))
    signarray = numpy.sign(derivative)
    crossing  = numpy.diff(signarray)
    indices   = crossing.nonzero()
    maximalist = []
    
    for i in indices[0]:
        # do not take into account the minima
        if crossing[i] > 0:
            continue

        if len(maximalist) == number_of_maxima:
            break

        f0 = freqs[i-1]
        f1 = freqs[i+1]

        infodict = scipy.optimize.brent(
                lambda f: 1./numpy.abs(CalculateImpedance(inst,f,T)),
                brack=(f0,f1),full_output=True)

        fres = infodict[0][0]
        zres = 1./infodict[1][0]

        # check that it is a maxima
        if zres > 1.:
            maximalist.append((fres, zres))

    return maximalist
