# coding: latin-1
# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Radiation.py - Classes defining the radiation impedance models

"""

__all__ = ['UnflangedOpenEnd','FlangedOpenEnd','IdealOpenEnd','ClosedEnd', 
        'SaxophoneBell']

import os
import math

import numpy
from numpy import log, sqrt

import scipy.integrate
from scipy.integrate import quad

import scipy.special
from scipy.special import jn_zeros, j1

import scipy.interpolate
from scipy.interpolate import UnivariateSpline

import wiat
from wiat.util import air_properties, transform_R2Z

# in this file we can add radiation models for standard horns
# possibly pre-calculated from BEM or MM models.
#
# Data should then be loaded from a file and impedance interpolated
# for tha values we need

class UnflangedOpenEnd():
    """ Unflanged open end radiation impedance models
    
	References:

            1. J. Dalmont and C.J. Nederveen,
               "Radiation impedance of tubes with different flanges:
	       numerical and experimental investigations ,"
               Journal of Sound and Vibration,  vol. 244, pp. 505-534, 2001.
            2. R. Causs, J. Kergomard, and X. Lurton,
               "Input impedance of brass musical instruments - 
	       Comparaison between experimental and numerical models,"
               J. Acoust. Soc. Am.,  vol. 75, pp. 241-254, 1984.
            3. H. Levine and J. Schwinger,
               On the radiation of sound from an unflanged circular pipe.
               Phys. Rev., 73(4), pp. 383-406, 1948.

        TODO: implement the exact solution from Levine and Schwinger
        
    """

    def __init__(self,d,model='Dalmont_Nederveen',m=1.):
        """ Radiation from an unflanged open pipe (Levine and schwinger)
            
            @param d: diameter of the open end
            @param m: multiplication factor (to take care of sperical wave, 
                approximation from Causse)
        """
        self.a = d/2.
        self.m = m
        self.model = model
        self.Z = {'Dalmont_Nederveen': self.Z_Dalmont_Nederveen,
                  'Causse': self.Z_Causse,
                  'Levine_Schwinger': self.Z_Levine_Schwinger}
        
    def __call__(self,f,T):
        return self.Z[self.model](f,T)

    def TM(self,f,T):
        return numpy.array([[self.Z[self.model](f,T)],
                            [numpy.ones(numpy.size(f))]],dtype=complex)

    def Z_Dalmont_Nederveen(self,f,T):
        """ Unflanged pipe radiation impedance approximation (ka < 3.5) from 
            Dalmont and Nederveen

            R = -|R0|exp(-2j*ka*delta), where delta is the frequency dependant 
            length correction
           
            See reference [1] page 509
        """
        air = air_properties(T)
        ka = 2*numpy.pi*f*self.a/air['c']
        R0 = (1 + 0.2*ka - 0.084*ka**2)/(1 + 0.2*ka + (0.5 - 0.084)*ka**2)
        delta = 0.6113 * ((1+0.044*ka**2)/(1+0.19*ka**2) -
                           0.02*(numpy.sin(2*ka))**2 )
        R = - R0 * numpy.exp( -2 * 1j * ka * delta)
        Zl = transform_R2Z(R)
        return self.m*Zl
    
    def Z_Causse(self,f,T):
        """ Unflanged pipe radiation impedance approximation (ka < 1.5)
            from Causse & al.
            See [2]
        """
        air = air_properties(T)
        ka = 2*numpy.pi*f*self.a/air['c']
        Zl = 0.25*(ka)**2 + 0.0127*(ka)**4 + 0.082*(ka)**4 * numpy.log(ka) -\
             0.023*(ka)**6 + 1j * (0.6113*(ka) - 0.036*(ka)**3 + \
             0.034*(ka)**3 * numpy.log(ka) - 0.0187*(ka)**5)
        return Zl

    def Z_Levine_Schwinger(self,f,T):
        air = air_properties(T)
        ka = 2*numpy.pi*f*self.a/air['c']

        #z = ka
        
        #int 2:  Eq. (VI.4) of Levine and Schwinger
        s = numpy.empty((len(ka)))
        for i,z in enumerate(ka):
            s[i] = scipy.integrate.quad(
                       lambda x,z: log(1/(2*scipy.special.i1(x)*\
                            scipy.special.k1(x)))/(x*scipy.sqrt(x**2 + z**2)),
                       0,20,(z,))
            
        #int 3:  Eq. (VI.5) of Levine and Schwinger
#        lambda x,z: numpy.arctan( scipy.special.k1(x)/(numpy.pi*scipy.special.i1(x)) ) * \
#               (1 - z/(numpy.sqrt(z**2 + x**2))) / x
        #int 4:  Eq. (VI.4) of Levine and Schwinger
#        lambda x,z:  nump.log(numpy.pi*numpy.abs(scipy.special.j1(x))*\
#                              numpy.sqrt(scipy.special.j1(x)**2 + scipy.special.y1(x)**2))/\
#                              (x*numpy.sqrt(z**2 - x**2));

class FlangedOpenEnd():
    def __init__(self,d,model='Dalmont_Nederveen'):
        
        self.a = d/2
        self.model = model
        self.Z = {'Dalmont_Nederveen': self.Z_Dalmont_Nederveen}

    def __call__(self,f,T):
        return self.Z[self.model](f,T)

    def TM(self,f,T):
        return numpy.array([[self.Z[self.model](f,T)],[numpy.ones(numpy.size(f))]],dtype=complex)

    def Z_Dalmont_Nederveen(self,f,T):
        """ Flanged pipe radiation impedance approximation (ka < 3.5) 
            from Dalmont and Nederveen

            R = -|R0|exp(-2j*ka*delta), where delta is the frequency dependant
            length correction

            Z = (1+R)/(1-R)
            See [1] page 509
        """
        air = air_properties(T)
        k = 2*numpy.pi*f/air['c']
        ka = k * self.a
        Rinf = (1 + 0.323*ka - 0.077*ka**2)/(1 + 0.323*ka + (1 - 0.077)*ka**2)
        #delta = 0.8216 * (1 + (0.77*ka)**2/(1+0.77*ka))**-1 # Dalmont

        delta = (0.82159-0.49*ka**2)/(1 -0.46*ka**3) # Norris
        
        R = - Rinf * numpy.exp( -2 * 1j * ka * delta)
        Zl = transform_R2Z(R)
        return Zl
    
    @staticmethod
    def Norris_Sheng_dbeta(s, x, i):
        return 2*s*j1(s)**2 / (sqrt(x**2-s**2) * (s**2-jn_zeros(1,i)**2))
                             
#    def Z_Norris_Sheng(self,f,T,number_of_modes = 15):
#        """
#            Norris, A. N., & Sheng, I. C. (1989).
#            Acoustic radiation from a circular pipe with an infinite flange.
#            Journal of Sound and Vibration, 135(1), 85-93.
#        """
#        
#        air = air_properties(T)
#        k = 2*numpy.pi*f/air['c']
#        ka = k * self.a
#
#        
#        alpha_00 = ka**-1 - j1(2*ka)/ka**2
#
#        M_00 = alpha_00 + 1.j/(ka) # eq. 9
#
#        # eq. 14, R as a function of ka
#        sum = 0.
#        for i in range(1,number_of_modes):
#            # eq. 13c
#            beta_i_1 = 1j*quad(Norris_Sheng_dbeta, 0, x, args=(ka,i)) + \
#                       quad(Norris_Sheng_dbeta,x,scipy.inf,args=(ka,i))
#            alpha_i0 = beta_i_1 # eq. 13b
#            if (ka > jn_zeros(1,i)):
#                xhi_i = numpy.sqrt(ka**2-jn_zeros(1,i)**2)/a
#            else:
#                xhi_i = 1j*numpy.sqrt(jn_zeros(1,i)**2-ka**2)/a
#                
#            B_i = A_i*xhi_i/(2*k) # eq. 12
#            sum = sum + alpha_i0*B_i
#        R = -1 + 2*(alpha_00 - sum)/M_00



class FlangedPiston():
	pass

class UnflangedPiston():
	pass

class IdealOpenEnd():
    def __call__(self,f,T):
        return numpy.zeros(numpy.size(f))

    def TM(self,f,T):
        return numpy.array([[numpy.zeros(numpy.size(f))],
                            [numpy.ones(numpy.size(f))]],dtype=complex)    

class ClosedEnd():
    def __call__(self,f,T):
        return numpy.inf*numpy.ones(numpy.size(f))

    def TM(self,f,T):
        return numpy.array([[numpy.ones(numpy.size(f))],
                            [numpy.zeros(numpy.size(f))]],dtype=complex)    


class SaxophoneBell():
    """ Radiation model of a saxophone bell.

        It uses the multimodal calculation method to calculate the planar
        wave input impedance of the bell.
    """
    def __init__(self, d, theta):
        """
        @param d: diameter of the bell at the junction with the body (small end)
        the diameter of the large end is implicit.
        @param theta: opening angle of the bell at the junction with the body.
        This is usually the same at the body angle in order to have a continuity
        in the tangent.
        """
        self.a = d/2.


        # the calculations are calculated as a function of ka and interpolated
        # if there is a diameter change afterward
        self.kas = numpy.arange(0.,5.,0.1)

        # We use a representative dimension for the losses to be valid
        self.d0 = 0.1
        self.de = 0.15

        # there is no Unflanged model yet...
        self.Zr = wiat.MM.Radiation.RadiationFlanged(self.de)

        self.bell = wiat.MM.Waveguides.QuarterCircleHorn(
                self.d0,self.de,math.radians(theta),math.radians(80.))

        self.number_of_modes = 5

        self.backup_filename = os.path.join(os.path.dirname(__file__),
                'SaxophoneBell.mat')

    def __str__(self):
        return "SaxophoneBell :: %s, length = %.4f"%\
           (str(self.bell), self.bell.L)


    def __call__(self,f,T):
        air = wiat.util.air_properties(T)

        # these are the value of ka for which we want the impedance to be 
        # returned
        kas = 2*numpy.pi*f*self.a/air['c']


        # this will create the impedance vector for the pre-defined value of
        # ka (self.kas)
        Z = wiat.util.load_or_calculate(self.backup_filename, 
                               wiat.MM.Solver.SolveImpedance,
                               geometry=self.bell, Zr=self.Zr,
#                               ks=2*self.kas/self.d0
                               freqs=air['c']*self.kas/(self.d0*numpy.pi), 
                               air=air,
                               N=self.number_of_modes)

        spl_real = UnivariateSpline(self.kas, numpy.real(Z),s=0.0000001)
        spl_imag = UnivariateSpline(self.kas, numpy.imag(Z),s=0.0000001)

        return spl_real(kas) - 1j*spl_imag(kas)


    def TM(self,f,T):
         return numpy.array([[self(f,T)],[numpy.ones(numpy.size(f))]],
                 dtype=complex)


