# coding: latin-1
#
# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" Waveguides.py - Contains classes defining the transmission matrices for
    simple geometries and functions to generate them.

"""

__all__ = ['Cylinder','Cone','CurvedCone',
           'ConeApproximatedWithCylinders','EquationApproximatedWithCones',
	   'CreateCylindersAndConesFromList', 'Jump', 'ToneHole']

from scipy import special
from scipy.integrate import quad

import numpy
from numpy import size, sqrt, cosh, sinh, tanh, ones

import Radiation
from wiat.util import air_properties

def PropagationConstant(f,a,air,beta):
    """ Calculate the complex propagation constant in an acoustic waveguide

        See Allan D. Pierce p. 531-534

        @param f: frequency array
	@param a: tube radius
	@param air: Air instance
	@param beta: factor to take into account wall roughness
	@return: S{Gamma} = S{alpha} + 1j * S{omega} / v_S{phi}
	    where S{alpha} is the attenuation, S{omega} is 2S{pi}f and
	    v_S{phi} is the phase velocity
	    
    """
    omega = 2*numpy.pi*f
    
    # slightly modified version of eq. 10-5.8
    alpha = beta * (1/a) * sqrt(omega*air['mu']/(2*air['rho']*air['c']**2)) *\
            (1 + (air['gamma']-1)/air['nu'])

    # slightly modified version of eq. 10-5.10
    G = 1j*(omega/air['c']) + (1+1j)*alpha 
    return G


class Waveguides:
    " Base class for waveguides "
    def __init__(self):
        self.a1 = 0.
        self.L = 0.

    def get_input_surface(self):
        " Return the input surface of the section. "
        return numpy.pi * self.a1 ** 2
    
    def get_length(self):
        " return the length of the section. "
        return self.L

    def checkx(self,x):
        assert -1e-6 <= x <= self.L+1e-6,\
                "value of x outside the waveguide by %.4f"%(x-self.L)
    
class Cylinder(Waveguides):
    " Class representing a cylinder "
    def __init__(self,d,L):
        """ Cylinder transmission matrix calculation class
        
            d: diameter
            L: length
        """
        self.a1 = d/2.
        self.L = L
   
    def __str__(self):
        return "Cylinder(d=%.4f,L=%.4f)"%(2*self.a1,self.L)

    def radius(self,x):
        self.checkx(x)
        return self.a1
    
    def TM(self,f,T):
        " Transmission matrix of the complete cylinder "
        TM = self.subTM(f,T,0.,self.L)        
        return TM

    def subTM(self,f,T,xi,xf):
        """ Transmission matrix of a section of the cylinder
            f: array of frequencies
            T: temperature
            xi: beginning of the section
            xf: end of the section
        """
        assert 0. <= xi <= self.L,\
            "cylinder subsection out of range %f <= %f <= %f"%(0.,xi,self.L)
        assert xi <= xf <= self.L+0.00001,\
            "cylinder subsection out of range %f <= %f <= %f"%(xi,xf,self.L)
        L = xf - xi
        air = air_properties(T)
        Gamma = PropagationConstant(f,self.a1,air,1)
        sinhL = numpy.sinh(Gamma*L)
        coshL = numpy.cosh(Gamma*L)
        TM = numpy.array([[coshL,sinhL],[sinhL,coshL]])    
        return TM

class Cone(Waveguides):
    """ Truncated cone transmission matrix calculation class
   
        This is an implementation of the formulation presented in [1]

        References:

            1. Transfer matrix of conical waveguides with any geometric
	       parameters for increased precision in computer modeling, 
	       The Journal of the Acoustical Society of America,  
	       vol. 122, pp. EL179-EL184, November 2007.
    """

    def __init__(self,d0,de,L):
        """ Initialize the conical waveguide data structure

	@param d0: diameter at the beginning [m]
        @type d0: float
	@param de: diameter at the end [m]
        @type de: float
	@param L: length of the cone [m]
	@type L: float
            
             
        """
        
        self.a1 = d0/2.
        self.ae = de/2.
        self.L = L

        self.x_in  = self.a1 * L / ( self.ae - self.a1 )
	# distance from apex to beginning of cone

        self.x_out = self.x_in + L
	# distance from apex to end of cone
       
    def __str__(self):
        return "Cone(d0=%.4f,de=%.4f,L=%.4f)"%(2*self.a1,2*self.ae,self.L)

    def halfangle(self):
        " Return the half angle of the cone"
        return numpy.arcsin(self.a1/self.x_in)
    
    def radius(self,x):
        """ Radius of the cone at distance x from its beginning """
        self.checkx(x)
        return (self.x_in + x) * self.a1 / self.x_in
    
    def _k(self,f,z):
        """ Local propagation constant in the conical waveguide
            at distance z from the apex
        """
        self.checkx(z - self.x_in)
        a = self.radius(z - self.x_in)
        air = air_properties(self.T)
        beta = 1.
        return -1j*PropagationConstant(f,a,air,beta)
        
    def _kbarL(self,freqs,xi,xf):
        """ equivalent propagation constant multiply by length of cone
            see Kulik eq.8
        """
        kbL = numpy.empty(numpy.size(freqs),dtype=complex)
        for i,f in enumerate(freqs):
            # we need integrate separately real and imaginary part because "
            # quad" do not support complex numbers
            kbarLr = quad(lambda x: numpy.real(self._k(f,x)), xi, xf,
                          epsabs=1e-10,epsrel=1e-8)
            kbarLi = quad(lambda x: numpy.imag(self._k(f,x)), xi, xf,
                          epsabs=1e-10,epsrel=1e-8)
            kbL[i] = kbarLr[0] +1j*kbarLi[0]
        return kbL

    def _theta(self,f,z):
        " z is the distance from the cone apex "
        return numpy.arctan( self._k(f, z) * z )
        
    def TM(self,f,T):
        self.T = T
        TM = self.subTM(f,T,0.,self.x_out-self.x_in)
        return TM

    def subTM(self,f,T,xi,xf):
        self.T = T

        x_in  = self.x_in + xi
        x_out = self.x_in + xf
        
        theta_in = self._theta(f,x_in)
        t_in = 1 / numpy.sin(theta_in);

        theta_out = self._theta(f,x_out) 
        t_out = 1 / numpy.sin(theta_out);

        r = x_out / x_in
        kbL = self._kbarL(f,x_in,x_out)
        
        TM = numpy.array([[-r*t_out*numpy.sin(kbL-theta_out), 
                           1j*r*numpy.sin(kbL)],
                          [1j*r*t_in*t_out*numpy.sin(kbL+theta_in-theta_out),
                           r*t_in*numpy.sin(kbL+theta_in)]])
        return TM

class CurvedCone(Cone):
    """ Class representing a curved cone 

        The idea is to construct an equivalent straight cone.
        The theory behind this have to be verify.

        References:
        
            1. Nederveen, C. J. (1998).
               Influence of a toroidal bend on wind instrument tuning.
               JASA, 104(3), 1616-1626.

    """
    def __init__(self,d0,de,L,R):
	""" Initialize the CurvedCone data structure 
        
        @param d0: diameter at the beginning [m]
        @type d0: float
        @param de: diameter at the end [m]
        @type de: float
        @param L: length of the cone [m]
        @type L: float
        @param R: the radius of curvature of the section
        @type R: float
	"""
        B0 = d0/(2*R)
        Be = de/(2*R)
        B = (B0+Be)/2.  # mean curvature

        # Nederveen, 1998, equation 9
        rhoB_rho = B**2/2 / (1 - sqrt(1 - B**2))
        LB = L*sqrt(rhoB_rho)
        d0B = 2*sqrt( (d0/2)**2 * sqrt(1/rhoB_rho) )
        deB = 2*sqrt( (de/2)**2 * sqrt(1/rhoB_rho) )

        Cone.__init__(self,d0B,deB,LB)

def CreateCylindersAndConesFromList(list):
    """ Build a list of Cylinder or Cone from a list of tuples containing 
        position and diameters.

	Example::
	    >>> import wiat
	    >>> FluteBody = wiat.TM.CreateCylindersAndConesFromList(
	    ... [(0.,0.0188),(0.132,0.0188),(0.340,0.0152),
	    ...  (0.452,0.0135),(0.583,0.0112)])
	    >>> for s in FluteBody: print s
            Cylinder(d=0.0188,L=0.1320)
            Cone(d0=0.0188,de=0.0152,L=0.2080)
            Cone(d0=0.0152,de=0.0135,L=0.1120)
            Cone(d0=0.0135,de=0.0112,L=0.1310)

        @param list: a list of tuples (position,diameter) where the 
	    position is the distance along the centerline of the instrument 
	    from the mouthpiece.

    """
    sections = []
    p0,d0 = list[0]
    for (p,d) in list[1:]:
        if d==d0:
            sections.append( Cylinder(d, p-p0) )
        else:
            sections.append( Cone(d0,d,p-p0) )
        p0,d0 = p,d
    return sections

def ConeApproximatedWithCylinders(d0,de,L,N=200):
    c = Cone(d0,de,L)
    lc = L / float(N)
    cylinders = []
    dp = d0
    for i in range(N):
        d = 2*c.radius(lc*(2*i+1)/2) 
        # diameter of the cone at the center of the section

        cylinders.append( Jump(dp,d) )
        cylinders.append( Cylinder(d,lc) )
        dp = d
    cylinders.append( Jump(d, de) )
    return cylinders

def EquationApproximatedWithCones(equation,L,N=10):
    """ Discretization in conical segments of an equation giving the radius of 
        the waveguide as a function of its argument.
    
        @param equation: an equation of one argument. It will be evaluated from
           0 to L.
        @param L: the length of the waveguide. 
    """
    lc = L / float(N)
    cones = []
    d1 = 2*equation(0)
    for i in range(N):
        x = (i+1)*lc
        d2 = 2*equation(x)
        cones.append( Cone(d1,d2,lc) )
        d1 = d2
    return cones
    
class Jump(Waveguides):
    " Class representing a diameter jump "
    def __init__(self,d1,d2):
        """ Transfer Matrix of a diameter jump
            d1: diameter 1
            d2: diameter 2
        """
        self.a1 = d1/2.
        self.a2 = d2/2.
        self.L = 0.

    # we should add two more terms
    #
    # inertance (La) and inductance (Ra)
    # see [1] Morse and Ingard, Theorical Acoustics, McGraw-Hill, 1968.
    # p. 488

    def subTM(self,f,T,xi,xf):
        return self.TM(f,T)
        
    def TM(self,f,T):
        S1 = numpy.pi * self.a1**2
        S2 = numpy.pi * self.a2**2
        n = numpy.size(f)
        TM = numpy.array([[numpy.ones(n),numpy.zeros(n)],
                          [numpy.zeros(n),numpy.ones(n)*S2/S1]])    
        return TM

class ToneHole:
    """ This class implements the transmission matrix of tone holes 
        References:
            1. J. Dalmont et al.,
               "Experimental Determination of the Equivalent Circuit of an 
	       Open Side Hole: Linear and Non Linear Behaviour,"
               Acta Acustica united with Acustica,  vol. 88, pp. 567-575, 2002.

            2. D.H. Keefe,
               "Theory of the single woodwind tonehole,"
               J. Acoust. Soc. Am.,  vol. 72, pp. 676-687, 1982.

            3. D.H. Keefe,
              "Experiments on the single woodwind tonehole,"
              J. Acoust. Soc. Am.,  vol. 72, pp. 688-699, 1982.

            4. van Walstijn, M., & Campbell, M. (2003).
               "Discrete-time modeling of woodwind instrument bores using wave 
	       variables,"
               J. Acoust. Soc. Am., vol. 113(1), 575-585.

            5. Nederveen, C. J. (1998).
               "Acoustical Aspects of Woodwind Instruments.", revised edition.
               147 pages. Northern Illinois University Press.

    """
    def __init__(self,d,t,p,d_pad=0.,h_pad=0.):
        """ Initialization of a tonehole data structure
            @param d: diameter of the tonehole
            @param t: height of the tonehole (minimum)
            @param p: position of the center of the tonehole with respect to 
	    the beginning of the instrument.
        """
        self.b = d/2.
        self.t = t
        self.p = p

        self.r_pad = d_pad/2.
        self.h_pad = h_pad

        # the radiation condition for holes should be investigated
	# as an approximation, use either Unflanged or Flanged open end

        #self.ToneholeOpenEnd = Waveguides.UnflangedOpenEnd(self.b)
        self.ToneholeOpenEnd = Radiation.FlangedOpenEnd(self.b)

    def __str__(self):
	    return "ToneHole(d=%.4f,t=%.4f,p=%.4f,d_pad=%.4f,h_pad=%.4f)"%\
			    (2*self.b,self.t,self.p,2*self.r_pad,self.h_pad)

    def relocate(self,new_p):
        """ Change the position of the tonehole.
        @param new_p: new position of the center of the tonehole with respect 
        to the beginning of the instrument.
        """
        self.p = new_p

    def resize(self,new_d):
        """ Change the diameter of the tonehole.
        @param new_d: new diameter of the tonehole
        """
        self.b = new_d/2.
        self.ToneholeOpenEnd = Radiation.FlangedOpenEnd(self.b)

    def get_position(self):
        return self.p

    def delta(self,a):
        """ Ratio of tone hole radius to body radius
            a: body radius
            
            The tone hole radius was set at initialization
        """
        return self.b/a

    def area(self):
        """ Cross sectional area of the tonehole """
        return  numpy.pi * self.b**2
    
    def InnerLengthCorrection(self,delta):
        " equation 4 from [1] "        
        return self.b * ( 0.82 - 1.4*delta**2 + 0.75*pow(delta, 2.7) )

    def MatchingVolumeLengthCorrection(self,delta):
        """ correction for matching volume, valid for cylindical bores
            ([1] eq. 6) 
        """
        return (self.b * delta / 8) * (1 + 0.207*delta**3)

    def RadiationLengthCorrection(self):
        """ from [5]
        """
        #Rc : outer radius of the hole (flange)
        #rd : extra length correction due to the key hanging above the hole
        
        if self.r_pad < self.b:
            td = 0.
        else:
            td = self.PadLengthCorrection()
            
        Rc = self.b # zero thickness hole...
        return self.b*(0.831 - 0.135*(self.b/Rc) -0.073*(self.b/Rc)**4) + td

        # we might ass eq. 9 Dalmont [1], for infinite flange

    def PadLengthCorrection(self):
        " Nederveen (1998) eq. 38.6 "

        # when there is no pad, radius is set to zero
        if self.r_pad == 0.:
            return 0.
        
        # the range of validity of this equation is
        if not (1. <= (self.r_pad/self.b) <= 4.):
            print "Warning, the r_pad/b ratio is out of range: %.2f"%\
                    (self.r_pad/self.b)
            
        tp = 0.613*self.b*((self.r_pad/self.b)**0.18 * \
                (self.r_pad/self.h_pad)**0.39 - 1)

        if tp < 0.: tp = 0.
        
        return tp
        
#    def ApproximateOpenShuntImpedance(self,f,a,T):
#        air = Air(T)
#        te_o = self.t + self.MatchingVolumeLengthCorrection(self.delta(a)) + \
        #self.RadiationLengthCorrection()
#        M = air.rho*te_o/self.area()
#        Zs_o = 1j*2*pi*f*M
#        return Zs_o
#    def ApproximateClosedShuntImpedance(self,f,a,T):
#        air = Air(T)
#        te_c = self.t + self.MatchingVolumeLengthCorrection(self.delta(a))
#        C = self.area()*te_c / (air.rho*air.c**2)
#        Zs_c = 1./(1j*2*pi*f*C)
#        return Zs_c
    
    def OpenShuntImpedance(self,f,a,T):
        """
        formulation with losses, etc.
        reference [1]
        """
        air = air_properties(T)
        k = 2*numpy.pi*f/air['c']
        Gamma = PropagationConstant(f,self.b,air,1)
        
        L = self.t + self.MatchingVolumeLengthCorrection(self.delta(a)) +\
                self.PadLengthCorrection()
        
        #print "t = %.4f, tm = %.4f, tp = %.4f, ti = %.4f"%\
                #(self.t,self.MatchingVolumeLengthCorrection(self.delta(a)),
                #self.PadLengthCorrection(),
                #self.InnerLengthCorrection(self.delta(a)))

        Zl = self.ToneholeOpenEnd(f,T)
        Zo = (Zl * cosh(Gamma*L) + sinh(Gamma*L)) / \
                (Zl * sinh(Gamma*L) + cosh(Gamma*L))
        Zi = 1j * k * self.InnerLengthCorrection(self.delta(a))
        Zc_hole = air['rho'] * air['c'] / self.area()
        return Zc_hole*(Zi+Zo)
    
    def ClosedShuntImpedance(self,f,a,T):
        """ impedance of a closed tone hole """
        air = air_properties(T)
        Gamma = PropagationConstant(f,self.b,air,1)
        L = self.t + self.MatchingVolumeLengthCorrection(a)
        # should we take into account the impedance of the pad, 
        # which is made of lether ?
        # we are currently assuming it has infinite impedance
        
        Zc_hole = air['rho'] * air['c'] / self.area()
        Zclosed = Zc_hole / tanh(Gamma*L)
        return Zclosed

    def SeriesLengthCorrection(self,delta):
        """ inertance
        
            see ref [1] eq. 2, 3 and 4
        """
        #t_a = -0.28 * self.b * delta**2 # eq. 2
        #t_a = -(0.37 - 0.087*delta) * self.b * delta**2 # eq.3

        t_a = -self.b * delta**2 * (2.72+0.540*delta+0.285*delta**2)**-1 
        # eq. 36 ref [4]

        return t_a

    
    def SeriesImpedance(self,f,a,T):
        """ Series impedance of the tone hole
        """
        air = air_properties(T)
        k = 2*numpy.pi*f/air['c']
        Zc_hole = air['rho'] * air['c'] / self.area()
        return -1j * Zc_hole * k * self.SeriesLengthCorrection(self.delta(a))
    
    def TM(self,f,a,T,state=0):
        """ Transmission matrix of the tone hole
        
            f: frequency array
            a: radius of the bore at tonehole location p
            T: temperature
            state: open or closed
        """

        assert a > self.b, "Bore smaller than tonehole (a = %.4f, b = %.4f)"%\
                (a,self.b)

        Za = self.SeriesImpedance(f,a,T) # same for open and closed hole
        if state == 1: # open
            Zs = self.OpenShuntImpedance(f,a,T)
        else: # closed
            Zs = self.ClosedShuntImpedance(f,a,T)
            
        
        # note: in our notation, using P and ZU as the two variables,
        # the matrices are made
        # of adimensional impedances (normalized with the characteristic 
        # impedance Zc)
        air = air_properties(T)
        Zc_tube = air['rho'] * air['c'] / (numpy.pi * a**2)
        TM = numpy.array([ [  ones(size(f)), (Za/Zc_tube)*ones(size(f))],
                           [Zc_tube/Zs, ones(size(f))] ])
        return TM        
