""" Saxophones.py - Definition of saxophone instruments

    It includes a Selmer Super Action Series 2 alto.
    Tonehole and dimensions measurements done by Gary Scavone and Tristan Matthews.
    Verification and some corrections to the measured data by Antoine Lefebvre

    Copyright (c) 2008 Antoine Lefebvre
    All rights reserved.
"""

import wiat
from wiat.TM import *

_L_ADAPTER = 0.0215 # length of the neck-body cylindrical adapter

_L_NECK = 0.172 + _L_ADAPTER 
# (the short cylinder part entering the main body is not included)

_L_TUBE = 0.475 - _L_ADAPTER
_L_ELBOW = 0.170

_D_NECK_IN  = 0.0125
_D_NECK_OUT = 0.0227

_D_BELL_OUT = 0.119

# measurement of the bell (this should be redone with more precise equipment)
_BELLDATA = [(0.,0.057),(0.06,0.063),(0.14,0.07),(0.16,0.073),
             (0.18,0.0765),(0.2,0.093),(0.205,0.119)]

_L_MOUTHPIECE_EQUIV = 0.0709
# equivalent mouthpiece length

_Mouthpiece = [Cylinder(_D_NECK_IN, _L_MOUTHPIECE_EQUIV)]

_Neck = [Cylinder(_D_NECK_IN, 0.0254),
         Cone(_D_NECK_IN, _D_NECK_OUT, _L_NECK-0.0254-_L_ADAPTER),
         Cylinder(_D_NECK_OUT, _L_ADAPTER) ]

_Body = [Cone(_D_NECK_OUT, 0.050, _L_TUBE)]

# Elbow = [Cone(0.050, 0.057, ElbowLength)]
# taking the curvature into account make a significant differences 
# (about 20 cents)
_Elbow = [CurvedCone(0.050, 0.057, _L_ELBOW, 0.0423)]

# the elbow is curved
# B = r0/R0 = 25.5/42.25 = 0.60 (small end)
# B = r0/R0 = 29.0/42.25 = 0.69 (large end)

_Bell = CreateCylindersAndConesFromList(_BELLDATA)


# bends: 

_Saxophone = _Neck + _Body + _Elbow + _Bell

# old body
#        Cone(0.057, 0.064, 0.069 + 0.0384),
#        Cone(0.064, 0.205, 0.14),
        

# diameter, th heigh min, max, position relative to neck input
offset = _L_NECK - _L_ADAPTER


# small function to convert tonehole positions from the side to the center
def __C(d,t,p,d_pad,h_pad):
    return (d,t,p+d/2.,d_pad,h_pad)

# 25 tone holes, including register holes
_ToneHoles = [
    ToneHole(*__C(0.00170, 0.0050, 0.0783, d_pad=0.003, h_pad=0.002)),
    ToneHole(*__C(0.01350, 0.0020, 0.0290+offset, d_pad=0.02, h_pad=0.002)),
    ToneHole(*__C(0.01350, 0.0090, 0.0450+offset, d_pad=0.02, h_pad=0.002)),
    ToneHole(*__C(0.00215, 0.0020, 0.0480+offset, d_pad=0.003, h_pad=0.002)),
    ToneHole(*__C(0.01330, 0.0020, 0.0580+offset, d_pad=0.02, h_pad=0.002)),
    ToneHole(*__C(0.01335, 0.0020, 0.0730+offset, d_pad=0.02, h_pad=0.002)),
    ToneHole(*__C(0.01320, 0.0015, 0.0930+offset, d_pad=0.02, h_pad=0.002)),
    ToneHole(*__C(0.00960, 0.0020, 0.1070+offset, d_pad=0.02, h_pad=0.002)),
    ToneHole(*__C(0.01940, 0.0020, 0.1270+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.01990, 0.0020, 0.1455+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.01510, 0.0020, 0.1625+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.01980, 0.0020, 0.1950+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.01990, 0.0020, 0.2010+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.02350, 0.0020, 0.2260+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.02550, 0.0020, 0.2650+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.02350, 0.0020, 0.3045+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.02770, 0.0020, 0.3360+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.02360, 0.0020, 0.3380+offset, d_pad=0.03, h_pad=0.002)),
    ToneHole(*__C(0.02970, 0.0020, 0.3850+offset, d_pad=0.035, h_pad=0.002)),
    ToneHole(*__C(0.03540, 0.0010, 0.4230+offset, d_pad=0.04, h_pad=0.002)),
    ToneHole(*__C(0.03290, 0.0020, 0.4860+offset, d_pad=0.04, h_pad=0.002)),
    ToneHole(*__C(0.02850, 0.0020, 0.0500+offset+_L_TUBE,
                 d_pad=0.035, h_pad=0.002)),
    ToneHole(*__C(0.02970, 0.0020,-0.0200+offset+_L_TUBE+_L_ELBOW,
                 d_pad=0.035, h_pad=0.002)),
    ToneHole(*__C(0.04010, 0.0020, 0.0120+offset+_L_TUBE+_L_ELBOW,
                 d_pad=0.05, h_pad=0.002)),
    ToneHole(*__C(0.03840, 0.0020, 0.0690+offset+_L_TUBE+_L_ELBOW,
                 d_pad=0.05, h_pad=0.002))]

# it would be nice to have a state machine giving the closed keys as a function of the key pressed

# there are 23 buttons to press (10 are 1 to 1 mapping) (4 are left small finger)

# Dictionary giving closed keys for a given note (standard fingering)
# 1: tonehole open
# 0: tonehole closed
# lowest tonehole to the right


_Fingerings = {
    'B flat 0':       [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
    'B 0':            [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1],
    'C 1':            [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1],
    'D flat 1':       [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1],
    'D 1':            [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1],
    'E flat 1':       [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1],
    'E 1':            [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,1],
    'F 1':            [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,1,1],
    'G flat 1':       [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,1,1],
    'G flat 1 (bis)': [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,0,1,0,1,1],
    'G 1':            [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,0,1,0,1,1], 
    'A flat 1':       [0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1,1,0,1,0,1,1],
    'A 1':            [0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'B flat 1':       [0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'B flat 1 (bis)': [0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,1,1,1,0,1,0,1,1],
    'B 1':            [0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'C 2 (bis)':      [0,0,0,0,0,0,0,0,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'C 2':            [0,0,0,0,0,0,0,0,1,0,0,0,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'D flat 2':       [0,0,0,0,0,0,0,1,1,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1], # all keys up
    'D 2':            [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1],
    'E flat 2':       [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1],
    'E 2':            [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,1],
    'F 2':            [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,1,1],
    'G flat 2':       [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,1,1],
    'G flat 2 (bis)': [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,0,1,0,1,1],
    'G 2':            [0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,0,1,0,1,1], 
    'A flat 2':       [0,0,0,1,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1,1,0,1,0,1,1],
    'A 2':            [1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'B flat 2':       [1,0,0,0,0,0,0,0,0,0,0,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'B flat 2 (bis)': [1,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,0,1,1,1,0,1,0,1,1],
    'B 2':            [1,0,0,0,0,0,0,0,0,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'C 3 (bis)':      [1,0,0,0,0,0,0,0,0,1,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'C 3':            [1,0,0,0,0,0,0,0,1,0,0,0,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'D flat 3':       [1,0,0,0,0,0,0,1,1,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1], # all keys up + register key
    'D 3':            [1,0,0,0,0,0,1,1,1,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'E flat 3':       [1,0,0,0,0,1,1,1,1,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'E 3 (palm)':     [1,0,0,0,1,1,1,1,1,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'E 3 (front)':    [1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,0,1,0,1,1],
    'F 3 (palm)':     [1,0,1,0,1,1,1,1,1,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1],
    'F 3 (front)':    [1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,0,1,0,1,1],
    'G flat 3 (palm)':[1,1,1,0,1,1,1,1,1,0,1,1,0,1,0,1,0,1,1,1,0,1,0,1,1] }

SelmerSerieII = Instrument(
    'Selmer serie II', _Saxophone, UnflangedOpenEnd(_D_BELL_OUT), 
    _ToneHoles, _Fingerings)

Selmer_wo_holes = Instrument(
    'Selmer serie II w/o toneholes', _Saxophone, 
    UnflangedOpenEnd(_D_BELL_OUT))

