import unittest
import os
import numpy
import scipy.io
import pylab

import wiat
from wiat.TM import *

#def load_or_calculate(filename, func, **args):
#    try:
#        f,Z = wiat.Data.load_from_mat(('f', 'Z'), filename)
#    except:
#        print 'No backup available, calculating...'
#        Z = func(**args)
#        scipy.io.savemat(filename, {'f': args['freqs'], 'Z': Z})
#    return Z

class TestHorns(unittest.TestCase):
    def setUp(self):
        self.f = numpy.arange(100.,5000.,10.)
        self.T = 14.307 # temperature at which c=340 (as for the BEM simulations)
        self.air = wiat.util.air_properties(self.T)
        
    def testQuarterCircleHorn(self):
        rin = 0.005 # input radius of the horn
        rhorn = 0.02 # radius of the circular horn
        din = 2*rin
        dend = 2*(rin+rhorn)
       
        fbem,Zbem = wiat.Data.load_from_mat(('f','Z'),
                wiat.BEMDIR+os.sep+'Z_horn4a.mat')
       
        fbem_flanged, Zbem_flanged =  wiat.Data.load_from_mat(('f','z'),
                wiat.BEMDIR+os.sep+'Z_horn4a_flanged.mat')

        k = 2*numpy.pi*fbem/self.air['c']
        ka = k*rin
        
        c = wiat.MM.Waveguides.QuarterCircleHorn(din,dend)

        Zr = wiat.MM.Radiation.RadiationFlanged(dend)

        Z1 = wiat.util.load_or_calculate('MM1_Horn4a.mat', 
                               wiat.MM.Solver.SolveImpedance,
                               geometry=c, Zr=Zr, freqs=fbem, air=self.air, 
                               N=1)

        Z5 = wiat.util.load_or_calculate('MM5_Horn4a.mat',
                               wiat.MM.Solver.SolveImpedance,
                               geometry=c, Zr=Zr, freqs=fbem, air=self.air, 
                               N=5)

        Z15 = wiat.util.load_or_calculate('MM15_Horn4a.dat', 
                                wiat.MM.Solver.SolveImpedance,
                                geometry=c, Zr=Zr, freqs=fbem, air=self.air, 
                                N=15)

        Horn = EquationApproximatedWithCones(
                lambda x: rin+rhorn-numpy.sqrt(rhorn**2 - x**2), rhorn, 120)
        
        If = Instrument("Quarter Circle Horn Unflanged", 
                       Horn, UnflangedOpenEnd(dend))

        Zf = wiat.util.load_or_calculate('TMf_Horn4a.mat',CalculateImpedance,
                               instrument=If, freqs=fbem, T=self.T)
        
        Iu = Instrument("Quarter Circle Horn Unflanged", 
                       Horn, FlangedOpenEnd(dend))

        Zu = wiat.util.load_or_calculate('TMu_Horn4a.mat',CalculateImpedance,
                               instrument=Iu, freqs=fbem, T=self.T)
        
        Zc = self.air['rho']*self.air['c']

        legends = ['cones flanged','cones unflanged','BEM unflanged',
                   'BEM flanged','MM flanged (1 mode)','MM flanged (5 modes)',
                   'MM flanged (15 modes)']
        fsize = (15,10)
        
        #pylab.figure(figsize=fsize)        
        #pylab.plot(ka,numpy.real(Zf),ka,numpy.real(Zu),ka,numpy.real(Zbem),ka,numpy.real(Zbem_flanged),ka,numpy.real(Z1),ka,numpy.real(Z5),ka,numpy.real(Z15))
        #pylab.legend(legends)
        #pylab.title('Real part of the input impedance')
        #pylab.xlabel('ka')
        #pylab.savefig('ReZ_horn4a')
        
        #pylab.figure(figsize=fsize)
        #pylab.plot(ka,numpy.imag(Zf),ka,numpy.imag(Zu),ka,numpy.imag(Zbem),ka,numpy.imag(Zbem_flanged),ka,-numpy.imag(Z1),ka,-numpy.imag(Z5),ka,-numpy.imag(Z15))
        #pylab.legend(legends)
        #pylab.title('Imaginary part of the input impedance')
        #pylab.xlabel('ka')
        #pylab.savefig('ImZ_horn4a')
        
        Rf = wiat.util.transform_Z2R(Zf)
        Ru = wiat.util.transform_Z2R(Zu)
        Rbem = wiat.util.transform_Z2R(Zbem)
        Rbem_flanged = wiat.util.transform_Z2R(Zbem_flanged)
        R1 = wiat.util.transform_Z2R(Z1)
        R5 = wiat.util.transform_Z2R(Z5)
        R15 = wiat.util.transform_Z2R(Z15)
        
        pylab.figure(figsize=fsize)
        pylab.plot(ka,numpy.abs(Rf),ka,numpy.abs(Ru),ka,numpy.abs(Rbem),ka,numpy.abs(Rbem_flanged),ka,numpy.abs(R1),ka,numpy.abs(R5),ka,numpy.abs(R15))
        pylab.legend(legends)
        pylab.title('Magnitude of the reflection coefficient')
        pylab.xlabel('ka')
        pylab.savefig('R_horn4a')
    
        
        tf = wiat.util.equivalent_length(Rf,k)
        tu = wiat.util.equivalent_length(Ru,k)
        tbem = wiat.util.equivalent_length(Rbem,k)
        tbem_flanged = wiat.util.equivalent_length(Rbem_flanged,k)
        t1 = wiat.util.equivalent_length(R1,k)
        t5 = wiat.util.equivalent_length(R5,k)
        #t15 = wiat.util.equivalent_length(R15,k)
        
        #pylab.figure(figsize=fsize)
        #pylab.plot(ka,tf/rin,ka,tu/rin,ka,tbem/rin,ka,tbem_flanged/rin,ka,-t1/rin,ka,-t5/rin,ka,-t15/rin)
        #pylab.ylim([0,3])
        #pylab.legend(legends)
        #pylab.title('Length corrections')
        #pylab.xlabel('ka')
        #pylab.savefig('L_horn4a')

        #pylab.show()
        
        #self.failUnlessAlmostEqual(t15[1], -tbem_flanged[1], 4)

    def testBesseHorn(self):
        pass
    def testCatenoidalHorn(self):
        pass
    def testExponentialHorn(self):
        pass
    


suite = unittest.TestSuite()
suite.addTest(TestHorns('testQuarterCircleHorn'))
unittest.TextTestRunner(verbosity=2).run(suite)

