import unittest
import numpy
from numpy import real, imag

import scipy.io

import pylab

import cProfile

from wiat.util import air_properties, transform_Z2R, equivalent_length
from wiat.TM import *
import wiat.MM

# The input impedance of a cylindrical tube is not perfectly identical
# with TM and MM simulations.
#
# The reflection coefficient of the tube shows oscillations that do not appear in the
# TM results, and that seems to depend upon the length of the tube
#
# MM results are alsways the conjugate of TM
#
def Load_or_Calculate(filename, f, func, args):
    try:
        d = scipy.io.loadmat(filename)
        Z = d['Z']
    except:
        print 'No backup available, calculating...'
        Z = func(*args)
        scipy.io.savemat(filename,{'f':f,'Z':Z})
    return Z

class TestMultiModal(unittest.TestCase):
    def setUp(self):
        self.f = numpy.arange(50,400,10)
        self.T = 14.307 # temperature at which c=340 (as for the BEM simulations)
        self.air = air_properties(self.T)
        
    def testFlangedCylinder(self):
        din = 0.0254
        L = 0.9

        k = 2*numpy.pi*self.f/self.air['c']
        ka = k*din/2
        
        c = wiat.MM.Waveguides.Cylinder(din,L)
        
        Z1 = Load_or_Calculate(
            'MM1_Cylinder',self.f, func=wiat.MM.Solver.SolveImpedance,
            args=(c,wiat.MM.Radiation.RadiationFlanged(din),self.f,self.air,1))
        Z5 = Load_or_Calculate(
            'MM5_Cylinder',self.f, func=wiat.MM.Solver.SolveImpedance,
            args=(c,wiat.MM.Radiation.RadiationFlanged(din),self.f,self.air,5))

        Zf = Load_or_Calculate('TMf_Cylinder',self.f,func=CalculateImpedance,
                               args=(Instrument("Test Cylinder Flanged", [Cylinder(din,L)], FlangedOpenEnd(din)),self.f,self.T))

        #print numpy.size(Z1), numpy.size(Z5), numpy.size(Zf), numpy.size(ka)
        
        pylab.figure()
        pylab.plot(ka,numpy.real(Z1),ka,numpy.real(Z5),ka,numpy.real(Zf))
        pylab.legend(('MM1','MM5','TM'))

        pylab.figure()
        pylab.plot(ka,-numpy.imag(Z1),ka,-numpy.imag(Z5),ka,numpy.imag(Zf))
        pylab.legend(('MM1','MM5','TM'))
        
        R1 = transform_Z2R(Z1)
        R5 = transform_Z2R(Z5)
        Rf = transform_Z2R(Zf)
        
        pylab.figure()
        pylab.plot(ka,numpy.abs(R1),ka,numpy.abs(R5),ka,numpy.abs(Rf))
        pylab.legend(('MM1','MM5','TM'))

        t1 = equivalent_length(R1,k)
        t5 = equivalent_length(R5,k)
        tf = equivalent_length(Rf,k)
        
        pylab.figure()
        pylab.plot(ka,-t1,ka,-t5,ka,tf)
        pylab.legend(('MM1','MM5','TM'))
        
        pylab.show()
        
suite = unittest.TestSuite()
suite.addTest(TestMultiModal('testFlangedCylinder'))
unittest.TextTestRunner(verbosity=2).run(suite)


#c = MultiModal.Cone(0.05,0.1,0.05)
#c = MultiModal.Cylinder(0.01,0.5)
#s1 = MultiModal.MultiModalSolver(c,1)
#s3 = MultiModal.MultiModalSolver(c,3)
#s6 = MultiModal.MultiModalSolver(c,6)
#s11 = MultiModal.MultiModalSolver(c,11)
#s17 = MultiModal.MultiModalSolver(c,17)
#freqs = numpy.arange(1800,2500,1)
#Z1 = s1.Zsolve(freqs)
#Z3 = s3.Zsolve(freqs)
#Z6 = s6.Zsolve(freqs)
#Z11 = s11.Zsolve(freqs)
#Z17 = s17.Zsolve(freqs)

#T = Temperature(25.,25.)
#T1 = Instrument("Test Ideal Open", [Cone(0.05,0.1,0.05), IdealOpenEnd()])
#Z = CalculatedImpedance("wo",T1,freqs,T,'noload')
#pylab.plot()

#pylab.plot(freqs,numpy.abs(Z1),freqs,numpy.abs(Z3),freqs,numpy.abs(Z6),freqs,numpy.abs(Z11),freqs,numpy.abs(Z.InputImpedance()))
#pylab.legend(['multimodal 1 mode','multimodal 3 modes','multimodal 6 modes','multimodal 11 modes','kulik'])

#pylab.plot(freqs,numpy.abs(Z1),freqs,numpy.abs(Z6),freqs,numpy.abs(Z11),freqs,numpy.abs(Z17),freqs,numpy.abs(Z.InputImpedance()))
#pylab.legend(['multimodal 1 mode','multimodal 6 modes','multimodal 11 modes','multimodal 17 modes','kulik'])

#ka=numpy.arange(0.1,10.,0.1)
#r = MultiModal.InfinitBaffleRadiation(0.02)
#Zs = r.Impedance(ka/r.radius)
#pylab.plot(ka,real(Zs[0][0]),ka,imag(Zs[0][0]),ka,real(Zs[1][1]),ka,imag(Zs[1][1]))#,ka,real(Zs[2][2]),ka,imag(Zs[2][2]))
           
#pylab.show()
