import unittest

import numpy
import pylab

import wiat
from wiat.TM import *

class TestWaveguides(unittest.TestCase):
    def setUp(self):
        self.f = numpy.arange(100,4600,1,dtype=float)
        self.T = 25.

    def testSplittedCone(self):
        """ Verify that the impedance of one long cone is identical to the same 
            cone splitted in two.

        """

        I1 = Instrument("Unflanged cone",
                        [Cone(0.01,0.03,0.3)], UnflangedOpenEnd(0.03))
        I2 = Instrument("Splitted unflanged cone",
                        [Cone(0.01,0.02,0.15), Cone(0.02,0.03,0.15)], 
                        UnflangedOpenEnd(0.03))
        Z1 = CalculateImpedance(I1,self.f,self.T)
        Z2 = CalculateImpedance(I2,self.f,self.T)
        c = CompareMaxima(CalculateMaxima(Z1,I1,self.f,self.T), 
                          CalculateMaxima(Z2,I2,self.f,self.T))
        for (f1, f2, cents, A1, A2, dB, r1, r2) in c:
            self.failUnless(cents < 1. and dB < 0.1,\
                    "Difference between maximas to big")

    def testSplittedCylinder(self):
        I1 = Instrument("Unflanged cylinder",
                        [Cylinder(0.02,0.3)], UnflangedOpenEnd(0.02))
        I2 = Instrument("Splitted unflanged cylinder",
                        [Cylinder(0.02,0.15), Cylinder(0.02,0.15)],
                        UnflangedOpenEnd(0.02))
        Z1 = CalculateImpedance(I1,self.f,self.T)
        Z2 = CalculateImpedance(I2,self.f,self.T)
        c = CompareMaxima(CalculateMaxima(Z1,I1,self.f,self.T),
                          CalculateMaxima(Z2,I2,self.f,self.T))
        for (f1, f2, cents, A1, A2, dB, r1, r2) in c:
            self.failUnless(cents < 1. and dB < 0.1,\
                    "Difference between maximas to big")
            
    def testComparison_Cone_CylinderApproximation(self):
        I1 = Instrument("Cone approximated with cylinders",
                        ConeApproximatedWithCylinders(0.01,0.03,0.3),
                        UnflangedOpenEnd(0.03))
        I2 = Instrument("Cone", [Cone(0.01,0.03,0.3)], UnflangedOpenEnd(0.03))
        Z1 = CalculateImpedance(I1,self.f,self.T)
        Z2 = CalculateImpedance(I2,self.f,self.T)
        c = CompareMaxima(CalculateMaxima(Z1,I1,self.f,self.T),
                          CalculateMaxima(Z2,I2,self.f,self.T))
        for (f1, f2, cents, A1, A2, dB, r1, r2) in c:
            self.failUnless(cents < 1. and dB < 0.1,\
                    "Difference between maximas to big")

    def testUnflangedOpenEnd(self):
        I1 = Instrument("Unflanged open end", [], UnflangedOpenEnd(0.1))
        Z1 = CalculateImpedance(I1,self.f,self.T)
        self.failUnlessAlmostEqual(numpy.max(numpy.real(Z1)), 1.032, 1)
        self.failUnlessAlmostEqual(numpy.max(numpy.imag(Z1)), 0.702, 1)

    def testClosedCylinder(self):
        L = 0.3
        I1 = Instrument("Closed cylinder", [Cylinder(0.02,L)], ClosedEnd())
        Z1 = CalculateImpedance(I1,self.f,self.T)
        m = CalculateMaxima(Z1,I1,self.f,self.T, number_of_maxima=1)
        self.failUnlessAlmostEqual(572.996, m[0][0], 3)

    def testUnflangedCylinder(self):
        L = 0.3
        I1 = Instrument("Unflanged cylinder", [Cylinder(0.02,L)], 
                        UnflangedOpenEnd(0.02))
        I2 = Instrument("Ideal open cylinder", [Cylinder(0.02,L)], 
                        IdealOpenEnd())
        Z1 = CalculateImpedance(I1,self.f,self.T)
        Z2 = CalculateImpedance(I2,self.f,self.T)
        pylab.plot(self.f,numpy.abs(Z1))
        pylab.plot(self.f,numpy.abs(Z2))
        pylab.legend(('unflanged','ideal open'))
        pylab.show()

suite = unittest.TestSuite()
suite.addTest(TestWaveguides('testUnflangedCylinder'))
suite.addTest(TestWaveguides('testClosedCylinder'))
suite.addTest(TestWaveguides('testSplittedCone'))
suite.addTest(TestWaveguides('testUnflangedOpenEnd'))
suite.addTest(TestWaveguides('testSplittedCylinder'))
suite.addTest(TestWaveguides('testComparison_Cone_CylinderApproximation'))

unittest.TextTestRunner(verbosity=2).run(suite)
