# coding: latin-1
#
# Copyright 2008 Antoine Lefebvre
#
# This file is part of "The Wind Instrument Acoustic Toolkit".
#
# "The Wind Instrument Acoustic Toolkit" is free software:
# you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# "The Wind Instrument Acoustic Toolkit" is distributed in the hope
# that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with "The Wind Instrument Acoustic Toolkit".
# If not, see <http://www.gnu.org/licenses/>.
#
# All rights reserved.

""" util.py - classes and functions that are not specific to the calculation
              methods.

"""

import numpy
import scipy

import wiat

def air_properties(T):
    """ Properties of air as a function of temperature

        The formula, valid for 10 Celcius around 26.85 C, comes from Table 1 
        of reference 1.
        
        References:
        
            1. Keefe, D. H. (1984). 
               Acoustical wave propagation in cylindrical ducts: Transmission 
               line parameter approximations for isothermal and nonisothermal 
               boundary conditions. J. Aoucst. Soc. Am., 75(1), 58-62.
    
        TODO: look for a more recent source to double check the validity

    """
    deltaT = T - 26.85
    properties = {
        # density   (kg/m^3)
        'rho': 1.1769    * (1 - 0.00335*deltaT),
        # viscosity (kg/m sec)
        'mu': 1.8460E-5 * (1 + 0.00250*deltaT),
        # specific heat ratio
        'gamma': 1.4017    * (1 - 0.00002*deltaT),
        # square root of Prandtl number : sqrt(Cp * mu / kappa)
        'nu': 0.8410    * (1 - 0.00020*deltaT),
        # free space sound speed (m/s)
        'c': 3.4723E+2 * (1 + 0.00166*deltaT)}
    return properties

def calculate_interval_in_cents(f1, f2):
    " Calculate the interval in cents between two frequencies. "
    return 1200. * numpy.log2(f2/f1)

def calculate_ratio_in_db(A1,A2):
    " Calculate the ratio in dB between two values. "
    return 20. * numpy.log10(A2/A1)

def transform_R2Z(R):
    " Convert reflection coefficient to normalized impedance "
    return (1+R)/(1-R)

def transform_Z2R(Z):
    " Convert normalized impedance to reflection coefficient "
    return (Z-1)/(Z+1)

def equivalent_length(R, k):
    """ Return the equivalent length of a complex reflection coefficient 
        R: reflection coefficient
        k: wavenumber
    """

    #return numpy.real( numpy.log(-R/numpy.abs(R)) / (-1j*2*k) )
    return numpy.angle(-R)/(-2*k)


def load_or_calculate(filename, func, **args):
    """ Function that will load a backup of the results if available,
    otherwise it calculates the function and save the results.
    """
    try:
        f,Z = wiat.Data.load_from_mat(('f', 'Z'), filename)
    except IOError,arg:
        print arg
        print 'No backup available (%s), calculating...'%filename
        Z = func(**args)
        scipy.io.savemat(filename, {'f': args['freqs'], 'Z': Z})
    return Z

def __create_equal_tempered_scale():
    scale = {}
    note_name = \
        ['Cx','Dx flat','Dx','Ex flat','Ex',\
         'Fx','Gx flat','Gx','Ax flat','Ax','Bx flat','Bx']

    # create the frequency dictionary for 8 octave
    for i in range(8):
        for j,note in enumerate(note_name):
            name = note.replace("x","%d"%i)
            scale[name] = 440*2**(-4+i+(j-9)/12.)
    return scale

EQUAL_TEMPERED_SCALE = __create_equal_tempered_scale()
